/*
 *
 * CLEX File Manager
 *
 * Copyright (C) 2001-2009 Vlado Potisk <vlado_potisk@clex.sk>
 *
 * CLEX is free software without warranty of any kind; see the
 * GNU General Public License as set out in the "COPYING" document
 * which accompanies the CLEX File Manager package.
 *
 * CLEX can be downloaded from http://www.clex.sk
 *
 */

#include "clexheaders.h"

#include <sys/stat.h>		/* stat() */
#include <errno.h>			/* errno */
#include <stdarg.h>			/* log.h */
#include <stdio.h>			/* fprintf() */
#include <stdlib.h>			/* free() */
#include <string.h>			/* strerror() */

#include "bookmarks.h"

#include "cfg.h"			/* cfg_num() */
#include "completion.h"		/* compl_text() */
#include "control.h"		/* control_loop() */
#include "directory.h"		/* dir_cmp() */
#include "edit.h"			/* edit_setprompt() */
#include "filepanel.h"		/* changedir() */
#include "filerw.h"			/* txtfile_open() */
#include "inout.h"			/* win_panel() */
#include "log.h"			/* msgout() */
#include "mbwstring.h"		/* convert2w() */
#include "panel.h"			/* pan_adjust() */
#include "util.h"			/* checkabs() */

static BOOKMARK *bm_list;			/* list of bookmarks */
static int bm_cnt = 0;				/* number of bookmarks */
static int bm_alloc = 0;			/* max. number of bookmarks */
static time_t bm_file_mod = 0;		/* last modification of the file */

static time_t
mod_time(const char *file)
{
	struct stat stbuf;

	return stat(file,&stbuf) < 0 ? 0 : stbuf.st_mtime;
}

static void
bm_save(void)
{
	int i, code;
	FILE *fp;

	if (user_data.nowrite) {
		msgout(MSG_W,"BOOKMARKS: Saving data to disk is prohibited");
		return;
	}

	msgout(MSG_HEADING,"BOOKMARKS: Writing the bookmarks file \"%s\"",
	  user_data.file_bm);

	errno = 0;
	fp = txtfile_fopen(user_data.file_bm);
	code = errno;
	if (fp == 0) {
		msgout(MSG_W,"Cannot open the bookmarks file for writing");
		if (code)
			msgout(MSG_W,"System error: %s",strerror(code));
		return;
	}
	fprintf(fp,"#\n# CLEX bookmarks file\n#\n");
	for (i = 0; i < bm_cnt; i++)
		fprintf(fp,"%s\n",USTR(bm_list[i].dir));
	if (txtfile_fclose(fp))
		msgout(MSG_W,"File write error occurred");
	else {
		msgout(MSG_HEADING,0);
		msgout(MSG_I,"BOOKMARKS: Data saved");
	}
	bm_file_mod = mod_time(user_data.file_bm);

}

/*
 * return value:
 *   0 = bookmarks loaded without problems
 *  -1 = failure, no or incomplete bookmarks were loaded
 */
static int
bm_read_from_file(void)
{
	int i, tfd, code;
	const char *line;
	FLAG corrupted;

	bm_cnt = 0;

	errno = 0;
	tfd = txtfile_open(user_data.file_bm,bm_alloc * 128);
	code = errno;
	if (tfd < 0 && tfd == FRW_ENOFILE)
		return 0;	/* missing optional file = ok */

	msgout(MSG_HEADING,"BOOKMARKS: Processing the bookmarks file \"%s\"",
	  user_data.file_bm);

	if (tfd < 0) {
		msgout(MSG_W,"File open error: %s",txtfile_strerror(tfd));
		if (code)
			msgout(MSG_W,"System error: %s",strerror(code));
		return -1;
	}

	code = txtfile_split(tfd,bm_alloc);
	if (code < 0) {
		msgout(MSG_W,"File parse error: %s",txtfile_strerror(code));
		if (code != FRW_ELINELIMIT) {
			txtfile_close(tfd);
			return -1;
		}
	}

	for (corrupted = 0, i = 0; (line = txtfile_line(tfd,i)); i++) {
		if (checkabs(line)) {
			us_copy(&bm_list[bm_cnt].dir,line);
			usw_convert2w(line,&bm_list[bm_cnt].dirw);
			bm_cnt++;
		}
		else if (!TSET(corrupted))
			msgout(MSG_W,"Invalid contents, file is corrupted");
	}
	txtfile_close(tfd);

	msgout(MSG_HEADING,0);
	return (code < 0 || corrupted) ? -1 : 0;
}

/*
 * bm_read_from_file() wrapper: reads the bookmarks if the bookmark file was modified
 *   or the 'force' flag is set 
 * return value: 0 file not read, -1 = read error, 1 = read success
 */
static int
bm_read(int force)
{
	time_t mod;
	int i, code;

	mod = mod_time(user_data.file_bm);
	if (mod == bm_file_mod && !force)
		/* file unchanged */
		return 0;
	
	code = bm_read_from_file();
	/* reset unused entries */
	for (i = bm_cnt; i < bm_alloc; i++) {
		us_reset(&bm_list[bm_cnt].dir);
		usw_reset(&bm_list[bm_cnt].dirw);
	}

	if (code == 0) {
		bm_file_mod = mod;
		return 1;
	}
	if (!user_data.nowrite)
		msgout(MSG_i,"This might help: "
		  "Main menu -> Bookmarks -> Organize bookmarks -> Save");
	return -1;
}

void
bm_reconfig(void)
{
	int i;

	if (bm_alloc) {
		for (i = 0; i < bm_cnt; i++) {
			us_reset(&bm_list[i].dir);
			usw_reset(&bm_list[i].dirw);
		}
		/* unused entries have been reset already */
		free(bm_list);
	}

	bm_cnt = 0;
	bm_alloc = cfg_num(CFG_B_SIZE);
	bm_list = emalloc(bm_alloc * sizeof(BOOKMARK));
	for (i = 0; i < bm_alloc; i++) {
		US_INIT(bm_list[i].dir);
		US_INIT(bm_list[i].dirw);
	}
	panel_bm_list.bm = panel_bm_org.bm = bm_list;

	bm_read(1);
}

int
bm_list_prepare(void)
{
	if (bm_read(0) == 1) {
		msgout(MSG_i,"New version of the bookmarks was loaded");
		panel_bm_list.pd->curs = panel_bm_list.pd->min;
	}

	panel_bm_list.pd->cnt = bm_cnt;
	if (panel_bm_list.pd->curs < 0)
		panel_bm_list.pd->curs = panel_bm_list.pd->top =  panel_bm_list.pd->min;
	/* otherwise leave the cursor where it is */

	panel = panel_bm_list.pd;
	textline = 0;
	return 0;
}

int
bm_org_prepare(void)
{
	panel_bm_org.pd->cnt = bm_cnt;
	panel_bm_org.pd->top = panel_bm_org.pd->curs = panel_bm_org.pd->min;

	panel = panel_bm_org.pd;
	textline = 0;
	return 0;
}

int
bm_edit_prepare(void)
{
	/* panel = panel_bm_org.pd; */
	textline = &line_tmp;
	edit_setprompt(textline,L"bookmark: ");
	edit_nu_putstr(USTR(bm_list[panel_bm_org.pd->curs].dirw));
	return 0;
}

void
cx_bm_org_addcwd(void)
{
	const char *dir;
	int i;

	dir = USTR(ppanel_file->dir);
	for (i = 0; i < bm_cnt; i++)
		if (strcmp(USTR(bm_list[i].dir),dir) == 0) {
			msgout(MSG_i,"Already bookmarked");
			panel_bm_org.pd->curs = i;
			pan_adjust(panel_bm_org.pd);
			win_panel_opt();
			return;
		}

	if (bm_cnt == bm_alloc) {
		msgout(MSG_W,"Bookmark list is full");
		return;
	}

	us_copy(&bm_list[bm_cnt].dir,dir);
	usw_convert2w(dir,&bm_list[bm_cnt].dirw);
	panel_bm_org.pd->curs = bm_cnt;
	panel_bm_org.pd->cnt = ++bm_cnt;
	pan_adjust(panel_bm_org.pd);
	win_panel();
}

void
cx_bm_list_chdir(void)
{
	if (changedir(USTR(bm_list[panel_bm_list.pd->curs].dir)) == 0)
		next_mode = MODE_SPECIAL_RETURN;
}

void
cx_bm_org_revert(void)
{
	bm_read(1);
	next_mode = MODE_SPECIAL_RETURN;
}

void
cx_bm_org_save(void)
{
	bm_save();
	next_mode = MODE_SPECIAL_RETURN;
}

static int
qcmp(const void *e1, const void *e2)
{
	return dir_cmp(USTR(((BOOKMARK *)e1)->dir),USTR(((BOOKMARK *)e2)->dir));
}

void
cx_bm_org_sort(void)
{
	qsort(bm_list,bm_cnt,sizeof(BOOKMARK),qcmp);
	win_panel();
}

void
cx_bm_org_edit(void)
{
	control_loop(MODE_BM_EDIT);
	win_panel_opt();
}

void
cx_bm_edit_enter(void)
{
	const wchar_t *dirw;

	if (*USTR(textline->line) != L'/') {
		msgout(MSG_w,"Directory name must start with a slash /");
		return;
	}
	dirw = USTR(textline->line);
	usw_copy(&bm_list[panel_bm_org.pd->curs].dirw,dirw);
	us_convert2mb(dirw,&bm_list[panel_bm_org.pd->curs].dir);
	next_mode = MODE_SPECIAL_RETURN;
}

void
cx_bm_edit_compl(void)
{
	if (compl_text(COMPL_TYPE_DIRPANEL) < 0)
		msgout(MSG_i,"COMPLETION: please type at least the first character");
}

void
cx_bm_org_up(void)
{
	int pos;

	if ((pos = panel_bm_org.pd->curs - 1) < 0)
		return;

	us_xchg(&bm_list[pos].dir,&bm_list[pos + 1].dir);
	usw_xchg(&bm_list[pos].dirw,&bm_list[pos + 1].dirw);

	panel_bm_org.pd->curs = pos;	/* curs-- */
	LIMIT_MAX(panel_bm_org.pd->top,pos);
	win_panel();
}

void
cx_bm_org_down(void)
{
	int pos;

	if ((pos = panel_bm_org.pd->curs + 1) == bm_cnt)
		return;

	us_xchg(&bm_list[pos].dir,&bm_list[pos - 1].dir);
	usw_xchg(&bm_list[pos].dirw,&bm_list[pos - 1].dirw);

	panel_bm_org.pd->curs = pos;	/* curs++ */
	LIMIT_MIN(panel_bm_org.pd->top,pos - disp_data.panlines + 1);
	win_panel();
}

void
cx_bm_org_del(void)
{
	int i, del;
	BOOKMARK x;

	del = panel_bm_org.pd->curs;

	panel_bm_org.pd->cnt = --bm_cnt;
	if (panel_bm_org.pd->curs == bm_cnt) {
		panel_bm_org.pd->curs--;
		pan_adjust(panel_bm_org.pd);
	}

	/* these are struct copy operations */
	x = bm_list[del];
	for (i = del; i < bm_cnt; i++)
		bm_list[i] = bm_list[i + 1];
	us_reset(&x.dir);
	usw_reset(&x.dirw);
	bm_list[bm_cnt] = x;

	win_panel();
}

void
cx_bm_org_new(void)
{
	int i, ins;
	BOOKMARK x;

	if (bm_cnt == bm_alloc) {
		msgout(MSG_W,"Bookmark list is full");
		return;
	}

	LIMIT_MIN(panel_bm_org.pd->curs,-1);
	ins = ++panel_bm_org.pd->curs;

	/* these are struct copy operations */
	x = bm_list[bm_cnt];
	for (i = bm_cnt; i > ins; i--)
		bm_list[i] = bm_list[i - 1];
	bm_list[ins] = x;
	panel_bm_org.pd->cnt = ++bm_cnt;

	usw_copy(&bm_list[ins].dirw,L"/");	/* initial value for editing */
	us_reset(&bm_list[ins].dir);		/* dir is null -> this is just a placeholder */
	cx_bm_org_edit();
	if (USTR(bm_list[ins].dir) == 0)
		/* dir is still null -> operation was canceled with ctrl-C */
		cx_bm_org_del();
}
