/*
 *
 * CLEX File Manager
 *
 * Copyright (C) 2001-2009 Vlado Potisk <vlado_potisk@clex.sk>
 *
 * CLEX is free software without warranty of any kind; see the
 * GNU General Public License as set out in the "COPYING" document
 * which accompanies the CLEX File Manager package.
 *
 * CLEX can be downloaded from http://www.clex.sk
 *
 */

/* text file read/write functions */

#include "clexheaders.h"

#include <sys/stat.h>		/* open() */
#include <errno.h>			/* errno */
#include <fcntl.h>			/* open() */
#include <stdio.h>			/* fopen() */
#include <stdlib.h>			/* free() */
#include <unistd.h>			/* fstat() */

#include "filerw.h"

#include "util.h"			/* emalloc() */

/* file read functions */

#define TFDESC_CNT	2		/* number of available descriptors (CLEX needs only 1) */
typedef struct {
	FLAG inuse;				/* 1 if valid */
	char *buff;				/* raw file data */
	size_t size;			/* file data size */
	char **line;			/* result of txtfile_split() */
	size_t linecnt;			/* number of lines */
} TFDESC;
static TFDESC tfdesc[TFDESC_CNT];

#define BADTFD(X) ((X) < 0 || (X) >= TFDESC_CNT || !tfdesc[X].inuse)
const char *
txtfile_strerror(int code)
{
	switch(code) {
	case FRW_EFULL:		return "Table of textfile descriptors is full";
	case FRW_EBAD:		return "Bad textfile descriptor";
	case FRW_ENOFILE:	return "File does not exist";
	case FRW_EOPEN:		return "Could not open the file";
	case FRW_EREAD:		return "Could not read the file";
	case FRW_ETYPE:		return "File is not a plain file";
	case FRW_ESAFE:		return "Unsafe file mode, file is world-writable";
	case FRW_ESIZELIMIT:return "File is too big (too many characters)";
	case FRW_ELINELIMIT:return "File is too big (too many lines)";
	default:			return "Unspecified error";
	}
}

/*
 * read text file into memory
 * output (success): txtfile descriptor >= 0
 * output (failure): error code < 0
 */
int
txtfile_open(const char *filename, size_t maxsize)
{
	int i, fd, tfd;
	struct stat stbuf;
	size_t filesize;	/* do not need off_t here */
	ssize_t rd;
	char *buff;

	if ( (fd = open(filename,O_RDONLY | O_NONBLOCK)) < 0)
		return errno == ENOENT ? FRW_ENOFILE : FRW_EOPEN;

	tfd = -1;
	for (i = 0; i < TFDESC_CNT; i++)
		if (!tfdesc[i].inuse) {
			tfd = i;
			break;
		}
	if (tfd == -1) {
		close(fd);
		return FRW_EFULL;
	}

	fstat(fd,&stbuf);	/* cannot fail with valid descriptor */
	if (!S_ISREG(stbuf.st_mode)) {
		close(fd);
		return FRW_ETYPE;
	}
	if ((stbuf.st_mode & S_IWOTH) == S_IWOTH) {
		close(fd);
		return FRW_ESAFE;
	}
	if (stbuf.st_size > maxsize) {
		close(fd);
		return FRW_ESIZELIMIT;
	}

	/* read into memory */
	filesize = stbuf.st_size;
	buff = emalloc(filesize + 1);	/* 1 extra byte for txtfile_split() */
	rd = filesize ? read_fd(fd,buff,filesize) : 0;
	close(fd);
	if (rd == -1) {
		free(buff);
		return FRW_EREAD;
	}

	tfdesc[tfd].inuse   = 1;
	tfdesc[tfd].buff    = buff;
	tfdesc[tfd].size    = filesize;
	tfdesc[tfd].line    = 0;
	tfdesc[tfd].linecnt = 0;
	return tfd;
}

int
txtfile_close(int tfd)
{
	if (BADTFD(tfd))
		return FRW_EBAD;

	free(tfdesc[tfd].buff);
	efree(tfdesc[tfd].line);
	tfdesc[tfd].inuse = 0;
	return 0;
}

/* split into lines, strip comments and empty lines */
int
txtfile_split(int tfd, size_t maxlines)
{
	size_t filesize;
	char ch, *buff, *line, *ptr;
	int ln;
	int comment;	/* -1 = don't know yet, 0 = no, 1 = yes */

	if (BADTFD(tfd))
		return FRW_EBAD;
	if (tfdesc[tfd].line)
		return 0;	/* everything is done already */

	buff = tfdesc[tfd].buff;
	filesize = tfdesc[tfd].size;

	/* terminate the last line if necessary */
	if (filesize && buff[filesize - 1] != '\n')
		buff[filesize++] = '\n';	/* this is the extra byte allocated in txtfile_open() */

	/* count all lines */
	for (ln = 0, ptr = buff; ptr < buff + filesize; ptr++)
		if (*ptr == '\t' || *ptr == '\r')
			*ptr = ' ';
		else if (*ptr == '\n' || *ptr == '\0') {
			*ptr = '\0';
			ln++;
	}
	LIMIT_MAX(ln,maxlines);

	/* assign line */
	tfdesc[tfd].line = emalloc(ln * sizeof(const char *));
	for (ln = 0, ptr = buff; ptr < buff + filesize; ) {
		line = ptr;
		for (comment = -1; (ch = *ptr++); ) {
			if (comment < 0) {
				if (ch == '#')
					comment = 1;
				else if (ch != ' ')
					comment = 0;
			}
		}
		if (comment == 0) {
			if (ln >= maxlines) {
				tfdesc[tfd].linecnt = maxlines;
				return FRW_ELINELIMIT;
			}
			tfdesc[tfd].line[ln++] = line;
		}
	}

	tfdesc[tfd].linecnt = ln;
	return 0;
}

int
txtfile_linecnt(int tfd)
{
	return BADTFD(tfd) ? -1 : tfdesc[tfd].linecnt;
}

/*
 * caution: the caller is allowed to overwrite the characters of the
 * returned string
*/
const char *
txtfile_line(int tfd, int lnum)
{
	if (BADTFD(tfd))
		return 0;
	if (tfdesc[tfd].line == 0 || lnum < 0 || lnum >= tfdesc[tfd].linecnt)
		return 0;
	return tfdesc[tfd].line[lnum];
}

/* file write functions */

FILE *
txtfile_fopen(const char *file)
{
	FILE *fp;

	umask(clex_data.umask | 022);
	fp = fopen(file,"w");
	if (fp == 0) {
		unlink(file);
		fp = fopen(file,"w");
	}
	umask(clex_data.umask);

	return fp;
}

int
txtfile_fclose(FILE *fp)
{
	FLAG errflag;

	errflag = ferror(fp) != 0;
	return fclose(fp) || errflag;
}
