/*
 *
 * CLEX File Manager
 *
 * Copyright (C) 2001-2009 Vlado Potisk <vlado_potisk@clex.sk>
 *
 * CLEX is free software without warranty of any kind; see the
 * GNU General Public License as set out in the "COPYING" document
 * which accompanies the CLEX File Manager package.
 *
 * CLEX can be downloaded from http://www.clex.sk
 *
 */

#include "clexheaders.h"

#include <errno.h>		/* errno */
#include <stdarg.h>		/* log.h */
#include <stdio.h>		/* printf() */
#include <string.h>		/* strerror() */
#include <unistd.h>		/* unlink() */

#include "opt.h"

#include "cfg.h"		/* cfg_num() */
#include "cmp.h"		/* cmp_saveopt() */
#include "filerw.h"		/* txtfile_open() */
#include "filter.h"		/* fopt_saveopt() */
#include "log.h"		/* msgout() */
#include "notify.h"		/* notif_saveopt() */
#include "sort.h"		/* sort_saveopt() */

/* limits to protect resources */
#define OPT_FILESIZE_LIMIT	150
#define OPT_LINES_LIMIT		 15

static FLAG changed = 0;			/* options have changed */

/* return value: 0 = ok, -1 = error */
static int
opt_read(void)
{
	int i, tfd, code;
	const char *line, *value;
	FLAG corrupted;

	errno = 0;
	tfd = txtfile_open(user_data.file_opt,OPT_FILESIZE_LIMIT);
	code = errno;
	if (tfd < 0 && tfd == FRW_ENOFILE)
		return 0;	/* missing optional file = ok */

	msgout(MSG_HEADING,"OPTIONS: processing the options file \"%s\"",
	  user_data.file_opt);

	if (tfd < 0) {
		msgout(MSG_W,"File open error: %s",txtfile_strerror(tfd));
		if (code)
			msgout(MSG_W,"System error: %s",strerror(code));
		return -1;
	}

	code = txtfile_split(tfd,OPT_LINES_LIMIT);
	if (code < 0) {
		msgout(MSG_W,"File parse error: %s",txtfile_strerror(code));
		if (code != FRW_ELINELIMIT) {
			txtfile_close(tfd);
			return -1;
		}
	}

	for (corrupted = 0, i = 0; (line = txtfile_line(tfd,i)); i++) {
		/* split VARIABLE and VALUE */
		if ( (value = strchr(line,'=')) == 0) {
			corrupted = 1;
			continue;
		}
		value++;
		if (strncmp(line,"COMPARE",7) == 0)
			code = cmp_restoreopt(value);
		else if (strncmp(line,"FILTER",6) == 0)
			code = fopt_restoreopt(value);
		else if (strncmp(line,"SORT",4) == 0)
			code = sort_restoreopt(value);
		else if (strncmp(line,"NOTIFY",6) == 0)
			code = notif_restoreopt(value);
		else
			code = -1;
		if (code < 0)
			corrupted = 1;
	}
	txtfile_close(tfd);

	if (corrupted) {
		msgout(MSG_W,"Invalid contents, file is outdated or corrupted");
		return -1;
	}
	msgout(MSG_HEADING,0);
	return 0;
}

void
opt_initialize(void)
{
	if (opt_read() < 0 && !user_data.nowrite) {
		/* automatic recovery */
		msgout(MSG_I,"Attempting to recover the options file (delete and save again)");
		unlink(user_data.file_opt);
		changed = 1;
		opt_save();
		msgout(MSG_I,changed ? "Attempt failed" : "File recovered");
	}
}

void
opt_changed(void)
{
	changed = 1;
}

/* this is a cleanup function (see err_exit() in control.c) */
void
opt_save(void)
{
	FILE *fp;

	if (!changed || user_data.nowrite)
		return;

	fp = txtfile_fopen(user_data.file_opt);
	if (fp == 0) {
		/*
		 * do not report errors to the user, because
		 * this action is not initiated by him/her
		 */
		msgout(MSG_NOTICE,"Cannot open options file for writing (%s)",
		  strerror(errno));
		return;
	}
	/*
	 * note: file output is block buffered, that's why a small file like this
	 * shall be written in a single write() call and opt_save() shall nor interfere
	 * with concurrently running CLEX processes writing the same file
	 */
	fprintf(fp,
	  "#\n# CLEX options file\n#\n"
	  "COMPARE=%s\n"
	  "FILTER=%s\n"
	  "SORT=%s\n"
	  "NOTIFY=%s\n",cmp_saveopt(),fopt_saveopt(),
	  sort_saveopt(),notif_saveopt());
	if (txtfile_fclose(fp))
		msgout(MSG_NOTICE,"Error writing options file (%s)",
		  strerror(errno));
	else
		changed = 0;
}
